////////////////////////////////////////////////////////////////////////////////
// BGBase
// Virtual class/memory tracer declaration and misc functions
// Written by Bjoern Ganster
// Developed since 2005
////////////////////////////////////////////////////////////////////////////////

#pragma once

#include "IL/config.h"

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include <time.h>

#include <iostream>
#include <exception>
#include <string>

#ifdef _WIN32
#include <Windows.h>
// At least for Windows' WriteFile, restrictions apply
#define ReadWriteBufMaxSize 64 * 1024
// Fix Windows' misbehaviour regarding min/max ...
#undef max
#undef min
#endif

#define storeClassID
#ifdef _DEBUG
#define SimpleMemoryCounters
#endif

////////////////////////////////////////////////////////////////////////////////
// Codepages

#define CP_ASCII 437
#define CP_UTF16_LE 1200
#define CP_UTF16_BE 1201
#define CP_UTF32 12000

////////////////////////////////////////////////////////////////////////////////
// Use C++11 features if supported

#if __cplusplus >= 201103 || (defined(_MSC_VER) && _MSC_VER >= 1900)
#define NOEXCEPT noexcept
#define OVERRIDE override 
#define FINAL final
//#define THREAD_LOCAL thread_local // no TLS in delay-lodaded-DLLs
#else
#define NOEXCEPT
#define OVERRIDE
#define FINAL
//#define THREAD_LOCAL // no TLS in delay-lodaded-DLLs

#endif

////////////////////////////////////////////////////////////////////////////////
// Operating system defines

#ifdef __APPLE__
typedef unsigned int uint;
#endif

#ifdef _WIN32
	typedef UINT BGKey;
	typedef wchar_t PathChar;
	//typedef char PathChar;
	const PathChar OSPathSep = L'\\';
	#define PathCharMod "%S"
	#define CP_SYSTEM CP_UTF16_LE
#else
	typedef char TCHAR;
	typedef uint BGKey;
	typedef char PathChar;
	typedef int64_t INT64;
	typedef u_int16_t WORD;
	typedef u_int32_t DWORD;
	const PathChar OSPathSep = '/';
	#define PathCharMod "%s"
	#define CP_UTF7 65000
	#define CP_UTF8 65001
	#define CP_SYSTEM CP_UTF8
#endif

////////////////////////////////////////////////////////////////////////////////

#define FILE_NO_BGBASE_CPP 1
#define FILE_NO_BGCPSTRING_H 2
#define FILE_NO_BGCPSTRING_CPP 3
#define FILE_NO_BGSTRINGTEMPLATES_H 4
#define FILE_NO_ILUT_WIN32_CPP 5
#define FILE_NO_IL_PALETTE_H 6
#define FILE_NO_IL_PALETTE_CPP 7
#define FILE_NO_IL_DEVIL_CPP 8
#define FILE_NO_IL_MANIP_CPP 9
#define FILE_NO_IL_STACK_CPP 10
#define FILE_NO_IL_API1_CPP 11
#define FILE_NO_IL_STATES_CPP 12
#define FILE_NO_IL_IO_CPP 13
#define FILE_NO_IL_ERROR_CPP 14
#define FILE_NO_IL_CONVERT_CPP 15

////////////////////////////////////////////////////////////////////////////////
// Put b into the range of a, c

template <class Type>
const Type& putInRange (const Type& min, const Type& val, const Type& max)
{
   if (val < min)
      return min;
   else if (val < max)
      return val;
   else
      return max;
}

////////////////////////////////////////////////////////////////////////////////
// Calculate minimum, maximum at the same time

template <class T>
void minmax (const T& a, const T& b, T& min, T& max)
{
   if (a < b) {
      T help = b; // Keep a copy in case of swap 
      min = a;
      max = help;
   } else {
      T help = a; // Keep a copy in case of swap 
      min = b;
      max = help;
   }
}

////////////////////////////////////////////////////////////////////////////////
// A simple integer storing the currently active operation - used for debugging
// crashes

class BGCPString;

class StackInfo FINAL {
public:
	bool cleanExit;

	StackInfo(const char* fileName, int fileNo, const char* functionName, int lineNo);
	~StackInfo();
	void setOp(const char* fileName, int fileNo, const char* functionName, int lineNo);

	int getFileNo() const
	{ return fileNo; }

	int getLineNo() const
	{ return lineNo; }

	StackInfo* getNext() const
	{ return next; }

private:
	int fileNo;
	int lineNo;
	StackInfo* next;
	StackInfo* previous;

	friend class StackTrace;
};

class StackTrace FINAL {
public:
	int mLastFileNo;
	int mLastLineNo;

	// Constructor
	StackTrace()
	: firstStackInfo(nullptr),
	  lastStackInfo(nullptr),
	  mLastFileNo(0),
	  mLastLineNo(0)
	{
	}

	// Destructor
	~StackTrace()
	{
	}

	void addStackFrame(StackInfo* info);

	void removeStackFrame();

	BGCPString toString() const;

	static const StackTrace* get();

	const StackInfo* getFirstinfo() const
	{ return firstStackInfo; }

private:
	StackInfo* firstStackInfo;
	StackInfo* lastStackInfo;
};

StackTrace* getStackTrace();
void logStackTrace();

////////////////////////////////////////////////////////////////////////////////

// This cannot currently work when building DLLs, since we are using classes,
// and we can only work with interfaces between a process and its DLLs ...
// Also thread_local is not compatible with delay-loaded-DLLs ...
//#define USE_STACK_INFO

#ifdef USE_STACK_INFO

#define SET_SCOPE StackInfo stackInfo (__FILE__, FILE_NUMBER, __FUNCTION__, __LINE__);

#define SET_OP stackInfo.setOp(__FILE__, FILE_NUMBER, __FUNCTION__, __LINE__);

#define RETURN stackInfo.cleanExit = true; return 

#define CONDITIONAL_THROW_WITH_BACKTRACE(condition, msg) \
if (condition) { \
	throw BGException(getStackTrace(), msg); \
}

#else
// Replacement definitions that do nothing ...
#define SET_SCOPE 

#define SET_OP 

#define RETURN return

#define CONDITIONAL_THROW_WITH_BACKTRACE(condition, msg) \
if (condition) { \
	throw BGException(getStackTrace(), msg); \
}
#endif

int getLastFileNo();
int getLastLineNo();
// Simple class to make sure that handles are closed 

#ifdef _WIN32

class Win32HandleCloser FINAL {
public:
	Win32HandleCloser(HANDLE handle = INVALID_HANDLE_VALUE);

	~Win32HandleCloser();

	void close();

	HANDLE get() const;

	bool valid();

	void operator=(HANDLE newVal);
private:
	HANDLE mHandle;
};

#endif

////////////////////////////////////////////////////////////////////////////////
// Cast data from D to C and throw an exception if data is outside C's limits

template <typename C, typename D>
C limitCheckCast(D data)
{
	if (data >= std::numeric_limits<C>::max()
	||  data <= std::numeric_limits<C>::min())
	{
		THROW_WITH_BACKTRACE("Limits exceeded");
	}

	return (C)(data);
}
