////////////////////////////////////////////////////////////////////////////////
// BGBase
// Virtual class/memory tracer declaration and misc functions
// Copyright Bjoern Ganster 
// Developed since in 2005
////////////////////////////////////////////////////////////////////////////////

#include "IL/BGBase.h"
#include "IL/devil_internal_exports.h"

#include <algorithm>
#include <iostream>

#define FILE_NUMBER FILE_NO_BGBASE_CPP

#ifndef _WIN32
#include "sys/time.h"
#include <sys/stat.h>
#endif

////////////////////////////////////////////////////////////////////////////////
// A simple integer storing the currently active operation - used for debugging
// crashes

#ifdef USE_STACK_INFO
// thread_local does not work with OpenMP?
StackTrace stackTrace;

StackInfo::StackInfo(const char* /*aFileName*/, int aFileNo, const char* /*aFunctionName*/, int aLineNo)
: fileNo (aFileNo),
  lineNo (aLineNo),
  next (NULL),
  previous(NULL),
  cleanExit(false)
{
	stackTrace.addStackFrame(this);
	stackTrace.mLastFileNo = aFileNo;
	stackTrace.mLastLineNo = aLineNo;
}

StackInfo::~StackInfo()
{
	// No way to figure out which image object is using this stack frame ... ?
	/*if (!cleanExit) {
		ilAddErrorA("Unclean exit at ");
		ilAddStackTrace();
	}*/
	stackTrace.removeStackFrame();
}

void StackInfo::setOp(const char* /*aFileName*/, int aFileNo, const char* /*aFunctionName*/, int aLineNo)
{
	fileNo = aFileNo;
	lineNo = aLineNo;
	stackTrace.mLastFileNo = fileNo;
	stackTrace.mLastLineNo = lineNo;
}

int getLastFileNo()
{ return stackTrace.mLastFileNo; }

int getLastLineNo() 
{ return stackTrace.mLastLineNo; }

const StackTrace* StackTrace::get() 
{
	return &stackTrace;
}

void StackTrace::addStackFrame(StackInfo* info)
{
	if (firstStackInfo == nullptr) {
		firstStackInfo = info;
		info->previous = nullptr;
	}

	if (lastStackInfo == nullptr) {
		lastStackInfo = info;
	} else {
		lastStackInfo->next = info;
		info->previous = lastStackInfo;
		lastStackInfo = info;
	}

	info->next = nullptr;
}

void StackTrace::removeStackFrame()
{
	if (firstStackInfo != nullptr && lastStackInfo != nullptr) {
		if (firstStackInfo == lastStackInfo) {
			firstStackInfo = nullptr;
			lastStackInfo = nullptr;
			mLastFileNo = 0;
			mLastLineNo = 0;
		} else {
			lastStackInfo = lastStackInfo->previous;
			lastStackInfo->next = nullptr;
			mLastFileNo = lastStackInfo->fileNo;
			mLastLineNo = lastStackInfo->lineNo;
		}
	}
}

void ilAddStackTrace(ILimage* image)
{
	const StackInfo* stackInfo = stackTrace.getFirstinfo();
	if (stackInfo != nullptr) {
		while  (stackInfo != nullptr) {
			ilAddErrorI(image, stackInfo->getFileNo(), 10);
			ilAddErrorA(image, ":");
			ilAddErrorI(image, stackInfo->getLineNo(), 10);
			ilAddErrorA(image, ", ");
		}
	} else {
		ilAddErrorA(image, " No stackinfo recorded.\n");
	}
}

StackTrace* getStackTrace()
{
	return &stackTrace;
}

#else

int gFileNo = 0, gLine = 0;
void setOp(int fileNo, int line)
{
	gFileNo = fileNo;
	gLine = line;
}

StackTrace* getStackTrace()
{
	return nullptr;
}

int getLastFileNo()
{
	return gFileNo;
}

int getLastLineNo() 
{
	return gLine;
}

/*BGCPString StackTrace::toString() const
{
	return BGCPString("No stack trace available");
}*/

void ilAddStackTrace(ILimage* image)
{
	ilAddErrorA(image, "Stackinfo: ");
	ilAddErrorI(image, gFileNo, 10);
	ilAddErrorA(image, ":");
	ilAddErrorI(image, gLine, 10);
	ilAddErrorA(image, ", ");
}


#endif

////////////////////////////////////////////////////////////////////////////////
// Simple class to make sure that handles are closed 

#ifdef _WIN32

Win32HandleCloser::Win32HandleCloser(HANDLE handle)
: mHandle (handle)
{
}

Win32HandleCloser::~Win32HandleCloser()
{
	close();
}

void Win32HandleCloser::close()
{
	SET_SCOPE;
	if (mHandle != INVALID_HANDLE_VALUE) {
		CloseHandle(mHandle);
		mHandle = INVALID_HANDLE_VALUE;
	}
	RETURN;
}

HANDLE Win32HandleCloser::get() const
{
	return mHandle;
}

bool Win32HandleCloser::valid()
{
	return (mHandle != INVALID_HANDLE_VALUE);
}

void Win32HandleCloser::operator=(HANDLE newVal)
{
	if (valid())
		CloseHandle(mHandle);
	mHandle = newVal;
}

#endif
