// LumpIO.h
// Implement IIO with pointer to data (called a Lump)
// Written by Bjrn Ganster in 2019 based on older code for X-Ways Software Technology AG

#pragma once

#include "Win32IO.h"
#include "PosixIO.h"

#include <IL/IIO.h>

class ConstLumpIO : public IIO {
public:
	// Constructor
	ConstLumpIO(const void* data = nullptr, size_t dataSize = 0, bool deleteData = false)
		: IIO(),
		mData((unsigned char*)data),
		mSize(dataSize),
		mOffset(0),
		mDeleteData(deleteData)
	{
	}

	// Destructor
	~ConstLumpIO() 
	{
		clear();
	}

	// Read lump from file
	ILboolean open(ILconst_string fileName);

	// Close operation, like for a file, clears contents here
	void close() OVERRIDE
	{
		clear();
	}

	// Clear contents
	void clear();

	ILint64 read(void* buffer, ILuint size, ILuint number)
	{
		if (buffer == NULL || mOffset > mSize)
			return 0;

		size_t readTotal = size * number;
		if (mOffset + readTotal > mSize)
			readTotal = mSize - mOffset;

		memcpy(buffer, mData + mOffset, readTotal);
		mOffset += readTotal;
		return readTotal / size;
	}

	ILint64 seek(ILint64 offset, ILuint mode) OVERRIDE
	{
		ILint64 newOffset = mOffset;
		switch (mode) {
		case IL_SEEK_SET:
			newOffset = offset;
			break;
		case IL_SEEK_CUR:
			newOffset += offset;
			break;
		case IL_SEEK_END:
		case IL_EOF:
			newOffset = mSize + offset;
			break;
		}

		mOffset = (size_t) newOffset;
		if (mOffset != newOffset) {
			return 1;
		} else
			return 0;
	}

	ILboolean eof() OVERRIDE
	{
		if (mOffset < mSize)
			return false;
		else
			return true;
	}

	ILint getc() OVERRIDE
	{
		ILint c = IL_EOF;

		if (mOffset < mSize) {
			c = mData[mOffset];
			++mOffset;
		}

		return c;
	}

	ILint64 tell() OVERRIDE
	{
		return mOffset;
	}

	ILint putc(ILubyte c) OVERRIDE
	{
		// Lumps are currently read-only
		++mOffset;
		return c;
	}

	ILint64 write(const void* buffer, ILuint size, ILuint number) OVERRIDE
	{
		return 0; // Lumps are currently read-only
	}
private:
	const unsigned char* mData;
	size_t mSize;
	size_t mOffset;
	bool mDeleteData;
};

class WritableLumpIO : public IIO {
public:
	// Constructor
	WritableLumpIO(void* data = nullptr, size_t dataSize = 0, bool deleteData = false)
		: IIO(),
		mData((unsigned char*)data),
		mSize(dataSize),
		mOffset(0),
		mDeleteData(deleteData)
	{
	}

	// Destructor
	~WritableLumpIO() 
	{
		clear();
	}

	// Read file contents into memory for updating
	ILboolean open(ILconst_string fileName);

	void close() OVERRIDE
	{
		clear();
	}

	// Forget stored contents
	void clear();

	ILint64 read(void* buffer, ILuint size, ILuint number)
	{
		size_t readTotal = size * number;
		if (mOffset + readTotal < mSize && buffer != NULL) {
			memcpy(buffer, mData + mOffset, readTotal);
			mOffset += readTotal;
			return readTotal;
		}
		else
			return 0;
	}

	ILint64 seek(ILint64 offset, ILuint mode) OVERRIDE
	{
		ILint64 newOffset = mOffset;
		switch (mode) {
		case IL_SEEK_SET:
			newOffset = offset;
			break;
		case IL_SEEK_CUR:
			newOffset += offset;
			break;
		case IL_SEEK_END:
		case IL_EOF:
			newOffset = mSize - offset;
			break;
		}

		mOffset = (size_t) newOffset;
		if (mOffset != newOffset) {
			return 1; // integer overflow?
		} else
			return 0;
	}

	ILboolean eof() OVERRIDE
	{
		if (mOffset < mSize)
			return false;
		else
			return true;
	}

	ILint getc() OVERRIDE
	{
		ILint c = IL_EOF;

		if (mOffset < mSize) {
			c = mData[mOffset];
			++mOffset;
		}

		return c;
	}

	ILint64 tell() OVERRIDE
	{
		return mOffset;
	}

	ILint putc(ILubyte c) OVERRIDE
	{
		if (mOffset < mSize) {
			mData[mOffset] = c;
			++mOffset;
		}
		return c;
	}

	ILint64 write(const void* buffer, ILuint size, ILuint number) OVERRIDE
	{
		size_t writeTotal = size * number;
		if (mOffset + writeTotal < mSize && buffer != NULL) {
			memcpy(mData + mOffset, buffer, writeTotal);
			mOffset += writeTotal;
			return writeTotal;
		}
		else
			return 0;
	}
private:
	unsigned char* mData;
	size_t mSize;
	size_t mOffset;
	bool mDeleteData;
};
